# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

#!/usr/bin/python

#
# IMPORTS
#

from __future__ import print_function
import re
import argparse
import sys
import clingo
import os
from ..spec_parser    import    spec_parser
from ..program_parser import program_parser
from ..solver         import         solver
from ..utils          import        printer
from ..utils          import          utils
from .                import    clingo_help
import errno
import signal

#
# DEFINES
#

UNKNOWN        = "UNKNOWN"
ERROR          = "*** ERROR: (asprin): {}"
ERROR_INFO     = "*** Info : (asprin): Try '--help' for usage information"
ERROR_OPEN     = "<cmd>: error: file could not be opened:\n  {}\n"
ERROR_FATAL    = "Fatal error, this should not happen.\n"
ERROR_PARSING  = "parsing failed"
DEBUG          = "--debug"
TEST           = "--test"
HELP_PROJECT   = """R|: Enable projective solution enumeration,
  projecting on the formulas of the specification"""
HELP_HEURISTIC = """R|: Apply domain heuristics with value <v> and modifier <m>
  on formulas of the preference specification"""
HELP_DELETE_BETTER = """R|: After computing an optimal model,
  add a program to delete models better than that one"""
HELP_TOTAL_ORDER = """R|: Do not add programs for optimal models after the \
first one
  Use only if the preference specification represents a total order"""
HELP_GROUND_ONCE = """R|: Ground preference program only once \
(for improving a model)"""
HELP_CLINGO_HELP = ": Print {1=basic|2=more|3=full} clingo help and exit"
HELP_RELEASE_LAST = """R|: Improving a model, release the preference program \
for the last model 
  as soon as possible"""
HELP_NO_OPT_IMPROVING = """R|: Improving a model, do not use optimal models"""
HELP_VOLATILE_IMPROVING = """R|: Use volatile preference programs \
for improving a model"""
HELP_VOLATILE_OPTIMAL = """R|: Use volatile preference programs \
for optimal models"""
HELP_TRANS_EXT = """R|: Configure handling of extended rules \
for non base programs
  (<m> should be as in clingo --trans-ext option)"""
HELP_PREFERENCE_UNSAT = """R|: Use """ + utils.UNSATP + """ programs \
for optimal models"""
HELP_CONST_NONBASE = """R|: Replace term occurrences of <id> in non-base
  programs with <t>"""
HELP_IMPROVE_LIMIT = """R|: Improving a model, stop search after x conflicts,
  where x is <m> times the conflicts for the first model of the current iteration;
  add ',all' to consider the conflicts for all the models of the current iteration,
  add ',<min>' to search always for at least <min> conflicts, 
  and add ',reprint' to reprint at the end the optimal models"""

#
# VERSION
#

HERE = os.path.abspath(os.path.dirname(__file__))
META_FILE = os.path.join(HERE,"..","..","__init__.py")
with open(META_FILE,'r') as f:
    meta = f.read()
    meta_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", meta, re.M)
    if meta_match:
        VERSION = meta_match.group(1)
    else:
        raise RuntimeError("Unable to find __version__ string")


#
# MyArgumentParser
#

class MyArgumentParser(argparse.ArgumentParser):

    def print_help(self, file=None):
        if file is None:
            file = sys.stdout
        file.write("asprin version {}\n".format(VERSION))
        argparse.ArgumentParser.print_help(self, file)

    def error(self, message):
        raise argparse.ArgumentError(None,"In context <asprin>: " + message)

class SmartFormatter(argparse.RawDescriptionHelpFormatter):

    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()
        return argparse.RawDescriptionHelpFormatter._split_lines(self, text, width)


#
# class AsprinArgumentParser
#
class AsprinArgumentParser:


    clingo_help = """
Clingo Options:
  --<option>[=<value>]\t: Set clingo <option> [to <value>]

    """


    usage = "asprin [number] [options] [files]"


    epilog = """
Default command-line:
asprin --models 1

asprin is part of Potassco: https://potassco.org/labs
Get help/report bugs via : https://potassco.org/support
    """


    version_string = "asprin " + VERSION + """
Copyright (C) Javier Romero
License: The MIT License <https://opensource.org/licenses/MIT>"""


    def __init__(self):
        self.underscores  = 0
        self.__first_file = None
        self.__file_warnings = []

    def __update_underscores(self,new):
        i = 0
        while len(new)>i and new[i]=="_": i+=1
        if i>self.underscores: self.underscores = i

    def __add_file(self,files,file):
        abs_file = os.path.abspath(file) if file != "-" else "-"
        if abs_file in [i[1] for i in files]:
            self.__file_warnings.append(file)
        else:
            files.append((file,abs_file))
        if not self.__first_file:
            self.__first_file = file

    def __do_constants(self, alist):
        try:
            constants = dict()
            for i in alist:
                old, sep, new = i.partition("=")
                self.__update_underscores(new)
                if new is "":
                    raise Exception(
                        "no definition for constant {}".format(old)
                    )
                if old in constants:
                    raise Exception("constant {} defined twice".format(old))
                else:
                    constants[old] = new
            return constants
        except Exception as e:
            self.__cmd_parser.error(str(e))

    def __do_improve_limit(self, string):
        if string is None:
            return None
        try:
            out = [0,False,0,False]
            match = re.match(r'(\d+)(,all)?(,\d+)?(,reprint)?$',string)
            if not match:
                raise Exception("incorrect value for option --improve-limit")
            out[0] = int(match.group(1))
            if match.group(2) is not None:
                out[1] = True
            if match.group(3) is not None:
                out[2] = int(match.group(3)[1:])
            if match.group(4) is not None:
                out[3] = True
            return out
        except Exception as e:
            self.__cmd_parser.error(str(e))

    def run(self, args):

        # command parser
        _epilog = self.clingo_help + "\nusage: " + self.usage + self.epilog
        cmd_parser = MyArgumentParser(usage=self.usage, epilog=_epilog,
            formatter_class=SmartFormatter, #argparse.RawDescriptionHelpFormatter,
            add_help=False, prog="asprin")
        self.__cmd_parser = cmd_parser

        # Basic Options
        basic = cmd_parser.add_argument_group('Basic Options')
        basic.add_argument('--help', '-h', action='help',
                           help=': Print help and exit')
        basic.add_argument('--clingo-help',
                           help=HELP_CLINGO_HELP,
                           type=int, dest='clingo_help', metavar='<m>', 
                           default=0, choices=[0,1,2,3])
        basic.add_argument('--version', '-v', dest='version',
                           action='store_true',
                           help=': Print version information and exit')
        basic.add_argument('--print-programs', dest='print-programs',
                           help=': Print translated programs and exit',
                           action='store_true')
        basic.add_argument('--no-check', dest='check', 
                           help=": Skip syntax checks",
                           action='store_false')
        #basic.add_argument('-', dest='read_stdin', action='store_true',
        #                   help=argparse.SUPPRESS)
        basic.add_argument(TEST, dest='test', action='store_true',
                           help=': Run system tests')
        basic.add_argument('--stats', dest='stats', action='store_true',
                           help=': Print statistics')
        basic.add_argument('--quiet', '-q', dest='quiet', choices=[0,1,2],
                           metavar='<q>', type=int, default=0, 
                           help=': print {0=all|1=optimal|2=no} models')
        #basic.add_argument('--no-info', dest='no_info', action='store_true',
        #                   help=': Do not print basic information')
        basic.add_argument('--no-asprin-lib', dest='asprin-lib',
                           help=': Do not include asprin_lib.lp',
                           action='store_false')
        basic.add_argument('-c', '--const', dest='constants', 
                           action="append", help=argparse.SUPPRESS, default=[])
        basic.add_argument('--const-nb', dest='constants_nb', 
                           action="append", 
                           metavar="<id>=<t>",
                           help=HELP_CONST_NONBASE,
                           default=[])
        basic.add_argument(DEBUG, dest='debug', action='store_true',
                           help=argparse.SUPPRESS)
        basic.add_argument('--to-clingo', dest='to_clingo', 
                           action="append", help=argparse.SUPPRESS, default=[])

        # Solving Options
        solving = cmd_parser.add_argument_group('Solving Options')
        solving.add_argument('--models', '-n',
                             help=": Compute at most <n> models (0 for all)",
                             type=int, dest='max_models', metavar='<n>',
                             default=1)
        solving.add_argument('--non-optimal', dest='non_optimal', 
                             help=": Compute also non optimal models", 
                             action='store_true')
        solving.add_argument('--project', dest='project', help=HELP_PROJECT,
                             action='store_true')
        solving.add_argument('--solving-mode', dest='solving_mode', 
                             metavar="<arg>",
                             help=""": Run {normal|approx|heuristic} \
                                       solving mode""",
                             #help=argparse.SUPPRESS,
                             default="normal", 
                             choices=["normal", "heuristic", "approx"])
        solving.add_argument('--dom-heur', dest='cmd_heuristic',
                              nargs=2, metavar=('<v>','<m>'),
                              help=HELP_HEURISTIC)

        # Additional Solving Options
        solving = cmd_parser.add_argument_group('Additional Solving Options')
        solving.add_argument('--steps', '-s', 
                             help=": Execute at most <s> steps", type=int,
                             dest='steps', metavar='<s>', default=0)
        basic.add_argument('--clean-up', dest='clean_up', action='store_true',
                           help=argparse.SUPPRESS)
        solving.add_argument('--delete-better', dest='delete_better', 
                             help=HELP_DELETE_BETTER,
                             action='store_true')
        solving.add_argument('--total-order', dest='total_order', 
                             help=HELP_TOTAL_ORDER,
                             action='store_true')
        solving.add_argument('--ground-once', dest='ground_once', 
                             help=HELP_GROUND_ONCE,
                             action='store_true')
        solving.add_argument('--release-last', dest='release_last', 
                             help=HELP_RELEASE_LAST,
                             action='store_true')
        solving.add_argument('--no-opt-improving', dest='no_opt_improving', 
                             help=HELP_NO_OPT_IMPROVING,
                             action='store_true')
        solving.add_argument('--volatile-improving', dest='volatile_improving', 
                             help=HELP_VOLATILE_IMPROVING,
                             action='store_true')
        solving.add_argument('--volatile-optimal', dest='volatile_optimal', 
                             help=HELP_VOLATILE_OPTIMAL,
                             action='store_true')
        solving.add_argument('--pref-trans-ext', dest='trans_ext',
                             help=HELP_TRANS_EXT, metavar="<m>", default=None)
        solving.add_argument('--preference-unsat', dest='preference_unsat',
                             help=argparse.SUPPRESS,
                             #help=HELP_PREFERENCE_UNSAT,
                             action='store_true')
        solving.add_argument('--improve-limit',
                             metavar='<m>', dest='improve_limit',
                             help=HELP_IMPROVE_LIMIT)

        options, unknown = cmd_parser.parse_known_args(args=args)
        options = vars(options)

        # print version
        if options['version']:
            print(self.version_string)
            sys.exit(0)

        # separate files, number of models and clingo options
        options['files'], clingo_options = [], []
        for i in unknown:
            if i=="-":                                   
                self.__add_file(options['files'],i)
            elif (re.match(r'^([0-9]|[1-9][0-9]+)$',i)): 
                options['max_models'] = int(i)
            elif (re.match(r'^-',i)):                    
                clingo_options.append(i)
            else:                                        
                self.__add_file(options['files'],i)

        # when no files, add stdin
        # build prologue
        if options['files'] == []:
            self.__first_file = "stdin"
            options['files'].append(("-","-"))
        if len(options['files'])>1: 
            self.__first_file += " ..."
        prologue = "asprin version " + VERSION + "\nReading from " 
        prologue += self.__first_file

        # handle constants
        options['constants']    = self.__do_constants(options['constants'])
        options['constants_nb'] = self.__do_constants(options['constants_nb'])

        # handle improve_limit
        options['improve_limit'] = self.__do_improve_limit(
            options['improve_limit']
        )

        # statistics
        # if options['stats']:
        clingo_options.append('--stats')

        # return
        return options, clingo_options, self.underscores, prologue, \
               self.__file_warnings



#
# class Asprin
#
class Asprin:

    def __init__(self):
        self.control = None
        self.solver = None
        self.options = None

    def __update_constants(self,options,constants):
        for i in constants:
            if i[0] not in options['constants']:
                options['constants'][i[0]] = i[1]

    def __get_control(self,clingo_options):
        try:
            return clingo.Control(clingo_options)
        except Exception as e:
            raise argparse.ArgumentError(None,e.message)

    def run_wild(self, args):

        # arguments parsing 
        aap = AsprinArgumentParser()
        self.options, clingo_options, u, prologue, warnings = aap.run(args)

        # clingo help
        if self.options["clingo_help"] > 0:
            if self.options["clingo_help"] == 1:
                out = clingo_help.HELP_1
            if self.options["clingo_help"] == 2:
                out = clingo_help.HELP_2
            if self.options["clingo_help"] == 3:
                out = clingo_help.HELP_3
            printer.Printer().do_print(out)
            sys.exit(0)

        # create Control object
        self.control = self.__get_control(clingo_options)

        # print prologue and warnings
        print(prologue)
        for i in warnings:
            printer.Printer().warning_included_file(i)

        # load --to-clingo files
        for i in self.options["to_clingo"]:
            self.control.load(i)

        # specification parsing
        sp = spec_parser.Parser(u)
        programs, utils.underscores, base_constants, self.options['show'] = \
                                                 sp.parse_files(self.options)
        self.__update_constants(self.options, base_constants)

        # preference programs parsing
        program_parser.Parser(self.control, programs, self.options).parse() 

        # solving
        self.solver = solver.Solver(self.control)
        self.solver.set_options(self.options)
        self.solver.run()

    def signal_handler(self, signum, frame):
        if self.solver is not None:
            self.solver.end()
        elif self.control is not None:
            printer.Printer().print_stats(self.control, 0, True, 0, False,
                                          self.options['stats'])
        sys.exit(1)

    def run(self, args):
        # signals
        signal.signal(signal.SIGTERM, self.signal_handler)
        signal.signal(signal.SIGINT,  self.signal_handler)
        # try to run wild
        try:
            self.run_wild(args)
        # catch exceptions
        except argparse.ArgumentError as e:
            print(ERROR.format(str(e)),file=sys.stderr)
            print(ERROR_INFO,file=sys.stderr)
            sys.exit(1)
        except IOError as e:
            if (e.errno == errno.ENOENT):
                print(ERROR_OPEN.format(e.filename),file=sys.stderr)
                print(ERROR.format(ERROR_PARSING),file=sys.stderr)
            else: 
                print(ERROR.format(str(e)),file=sys.stderr)
            print(UNKNOWN,file=sys.stdout)
            sys.exit(65)
        except utils.SilentException as e:
            pass
        except utils.FatalException as e:
            print(ERROR.format(ERROR_FATAL),file=sys.stderr)
            print(UNKNOWN,file=sys.stdout)
            sys.exit(65)
        except Exception as e:
            print(ERROR.format(str(e)),file=sys.stderr)
            print(UNKNOWN,file=sys.stdout)
            sys.exit(65)


def main(args):
    if TEST in args:
        args.remove(TEST)
        from ..tests import tester
        tester.main(args)
    elif DEBUG in args:
        Asprin().run_wild(args)
    else:
        Asprin().run(args)





# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

#!/usr/bin/python
from __future__ import print_function
import os
import utils
import sys
import subprocess

PATH = os.path.dirname(os.path.realpath(__file__))
DIR = "--test-dir="
ALL = "--all"
OPTIONS = [
    [""],
    ["--delete-better"],
    ["--ground-once"],
    ["--release-last"],
    ["--no-opt-improving"],
    ["--volatile-improving"],
    ["--volatile-optimal"],
    #["--preference-unsat $asprin/mine/asprin_lib_unsat.lp"],
    ["--solving-mode=heuristic"],
    ["--solving-mode=heuristic --const-nb heuristic_aso=2"],
    ["--solving-mode=heuristic --const-nb heuristic_aso=2 --const-nb use_get_sequence=2 "],
    ["""--solving-mode=heuristic --const-nb heuristic_poset=2 --const-nb heuristic_aso=3 \
     --const-nb heuristic_pareto=2 --const-nb heuristic_and=2"""],
    ["--solving-mode=approx"],
    ["--solving-mode=approx --const-nb approx_aso=2"],
    ["--solving-mode=approx --const-nb approx_aso=2 --const-nb use_get_sequence=2 "],
    ["""--solving-mode=approx --const-nb approx_poset=2 --const-nb approx_aso=3 \
     --const-nb approx_pareto=2 --const-nb approx_and=2"""],
    ["--improve-limit=0,10"],
    ["--improve-limit=1,all,10,reprint"],
]

class cd:
    """Context manager for changing the current working directory"""
    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)

class Tester:
    
    def run(self, dir, options):
        errors, error = False, False
        for i in sorted(os.listdir(dir)):
            abs_i = os.path.join(dir,i)
            if os.path.isdir(abs_i):
                error = self.run(os.path.join(dir, i), options)
            elif str(abs_i)[-3:] == ".lp":
                with open(abs_i, 'r') as f:
                    test = utils.Test(f.read(), options)
                print("Testing {}...".format(abs_i))
                import tempfile
                tmp = tempfile.TemporaryFile()
                with cd(dir):
                    subprocess.call(test.command, stdout=tmp,
                                    stderr=subprocess.STDOUT, shell=True)
                    tmp.seek(0)
                    result = utils.Result(tmp.read())
                error = result.compare(test)
            if error:
                errors = True
        return errors

def main(args):
    path = PATH
    for i in args:
        if i.startswith(DIR):
            path = i[len(DIR):]
            args.remove(i)
            break
    if ALL in args:
        options = OPTIONS
    else:
        options = [args]
    for opt in options:
        print("Options = {}".format(opt))
        errors = Tester().run(path, opt)
        if errors:
            print("ERROR: There were errors in the tests")
        else:
            print("OK: All tests were successful")

if __name__ == "__main__":
    main()




# MIT License
# 
# Copyright (c) 2017 Javier Romero
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-

# this file should be reimplemented at some point...

from ..utils import utils


#
# defines 
#

# predicates
PREFERENCE         = "preference"   # arity 2 and 5
OPTIMIZE           = "optimize"     # arity 1
HOLDS              = "holds"        # arity 2
SAT                = "sat"          # arity 1
BF                 = "bf"           # arity 1
DOM                = "dom"          # arity 1
PREFERENCE_DOM     = "pref_dom"     # arity 1
GEN_DOM            = "gen_dom"      # arity 2
GEN_PREFERENCE_DOM = "gen_pref_dom" # arity 2

# terms
TRUE  = "true"
FALSE = "false"
ATOM  = "atom"
CMP   = "cmp"
NEG   = "neg"
AND   = "and"
OR    = "or"
NOT   = "not"
NAME  = "name"
FOR   = "for"
HASH_SEM = utils.HASH_SEM

# boolean formulas
BF_ENCODING = """
##sat(and(X,Y)) :- ##sat(X), ##sat(Y), ##bf(and(X,Y)).
##sat(or (X,Y)) :- ##sat(X),           ##bf(or (X,Y)).
##sat(or (X,Y)) :- ##sat(Y),           ##bf(or (X,Y)).
##sat(neg(X  )) :- not ##sat(X),       ##bf(neg(X  )).
##bf(X) :- ##bf(and(X,Y)).
##bf(Y) :- ##bf(and(X,Y)).
##bf(X) :- ##bf(or (X,Y)).
##bf(Y) :- ##bf(or (X,Y)).
##bf(X) :- ##bf(neg(X  )).
"""

# true atom
TRUE_ATOM = """
##true.
"""

# rule for pref_dom/1
PREF_DOM_RULE = """
##pref_dom(X) :- ##preference(X,_).
"""


#
# Global Functions
#

#
# Note: an ast is either None, a string, or a list
#

# Translate ast to string
def ast2str(ast):
    if ast == None:
        return ""
    if isinstance(ast,str): 
        return ast
    char, alist = "", ast
    if len(ast)==2 and ast[0]==HASH_SEM: # pooling
        char, alist = ";", ast[1]
    return char.join([ast2str(e) for e in alist])


# Translate body to string
def _body2str(i):
    out = []
    if isinstance(i,list) and len(i) > 0:
        if isinstance(i[0],str) and i[0] in { ATOM, TRUE, FALSE, CMP}:
            return [ ast2str(i[1]) ]
        for j in i:
            out += _body2str(j)
    return out

# TODO: CHECK
def body2str(i):
    return ", ".join(_body2str(i))

# get variables from ast
def get_vars(ast):
    if ast == None: return []
    if isinstance(ast,str):
        return [ast] if len(ast)>0 and (ast[0]>="A" and ast[0]<="Z") else []
    return [ i for l in map(get_vars,ast) for i in l ]


# has an ast any variables?
def has_vars(ast):
    if ast == None: return False
    if isinstance(ast,str):
        return (len(ast)>0 and (ast[0]>="A" and ast[0]<="Z"))
    for i in ast:
        if has_vars(i): return True
    return False


#
# abstract class for preference and optimize statements
#
class Statement:

    underscores = ""
    domains  = set()

    def __init__(self):
        self.number   = None
        self.name     = None
        self.type     = None
        self.elements = []
        self.body     = None


#
# preference statement
#
class PStatement(Statement):


    bfs = False # True if there are boolean formulas which are not literals


    def __create_dom_body(self, element):
        out = [k for i in element.sets for j in i for k in j.get_dom_atoms()]
        for k in element.cond:
            out += k.get_dom_atoms()
        Statement.domains.update([x[1] for x in out if x[1] is not None])
        return ", ".join(x[0] for x in out)


    def str(self):

        # underscores
        u = Statement.underscores

        # tostring
        name = ast2str(self.name)
        type = ast2str(self.type)

        # pref/2
        statement_body = body2str(self.body) if self.body is not None else ""
        arrow = " :- " if statement_body != "" else ""
        out  = u + PREFERENCE 
        out += "({},{}){}{}.\n".format(name, type, arrow, statement_body)

        # handle pooling
        for i in self.elements:
            if i.pooling:
                self.elements = [k for j in self.elements for k in j.unpool()]
                break

        # pref/5
        elem = 1
        for i in self.elements:

            set = 1
            # body
            if i.body is not None:
                body = body2str(i.body)
            else:
                body = self.__create_dom_body(i)
            if statement_body != "":
                if body != "": 
                    body += ", "
                body += statement_body
            arrow = " :- " if body != "" else ""

            # head sets
            for j in i.sets:
                for k in j:
                    out += u + PREFERENCE
                    s = "({},({},{},({})),{},{},{}){}{}.\n"
                    out += s.format(name, self.number, elem, ",".join(i.vars),
                                    set, k.str_body(), k.str_weight(), arrow,
                                    body)
                    out += k.str_holds(body)
                    out += k.str_bf   (body)
                    out += k.str_sat  (body)
                set += 1

            # condition set
            for k in i.cond:
                out += u + PREFERENCE
                s = "({},({},{},({})),{},{},{}){}{}.\n"
                out += s.format(name, self.number, elem, ",".join(i.vars), 0,
                                k.str_body(), k.str_weight(), arrow, body)
                out += k.str_holds(body)
                out += k.str_bf   (body)
                out += k.str_sat  (body)

            elem += 1
        #end for

        return out


#
# optimize statement
#
class OStatement(Statement):

    def str(self):
        statement_body = body2str(self.body) if self.body is not None else ""
        arrow = " :- " if statement_body != "" else ""
        out = Statement.underscores + OPTIMIZE 
        out += "({}){}{}.\n".format(ast2str(self.name), arrow, statement_body)
        return out


#
# program statement
#
class ProgramStatement(Statement):

    def str(self):
        return ast2str(self.name)


#
# preference element
#
class Element:


    def __init__(self):
        self.sets     = []
        self.cond     = []
        self.body     = None
        self.vars     = set()
        self.all_vars = set() # temporary variable
        self.pooling  = False


    def get_vars(self):
        vars = [ k for i in self.sets for j in i for k in j.get_vars() ]
        for j in self.cond:
            vars += j.get_vars()
        return set(vars)


    def unpool(self):
        if not self.pooling: return [self]
        sets = [[]]
        for i in self.sets: # sets
            set = [[]]
            for j in i:              # weighted elements
                wBodies = j.unpool() # a list of WBodies
                set = [x + [y] for x in set for y in wBodies]
            sets = [x + [y] for x in sets for y in set]
        cond = [[]]
        for i in self.cond: # cond
            wBodies = i.unpool() # a list of WBodies
            cond = [x + [y] for x in cond for y in wBodies]
        element_pairs = [(x,y) for x in sets for y in cond]
        elements = []
        for i in element_pairs:
            e = Element()
            e.sets = i[0]
            e.cond = i[1]
            e.body = self.body
            e.vars = e.get_vars()
            elements.append(e)
        return elements


#
# weighted body
#
class WBody:


    def __init__(self,weight,body,naming=False):
        self.weight      = weight
        self.body        = body  # list
        self.naming      = naming
        # reified boolean formula representing the body
        self.bf          = None  
        # extended atoms appearing in (boolean formulas which are not literals)
        self.atoms_in_bf = set() 
        self.analyzed    = False


    #
    # UNPOOLING
    #


    def get_vars(self):
        vars  = []
        vars += get_vars(self.weight)
        vars += get_vars(self.body)
        return vars


    def __unpool_term(self, term):
        if term == None:              return [None]
        if not isinstance(term,list): return [term]
        if len(term)==2 and term[0] == HASH_SEM:
            out = []
            for i in term[1]:
                out += self.__unpool_term(i)
            return out
        out = [[]]
        for i in term:
            out = [x + [y] for x in out for y in self.__unpool_term(i)]
        return out


    def unpool(self):
        weights = self.__unpool_term(self.weight)
        bodies  = self.__unpool_term(self.body)
        return [WBody(i,j,self.naming) for i in weights for j in bodies]


    #
    # BOOLEAN FORMULAS TRANSLATION
    #


    def __translate_ext_atom(self, atom):
        u = Statement.underscores
        if atom[0] == "true":
            out = ATOM + "(" + u + TRUE + ")"
            return out, u + TRUE
        elif atom[0] == "false":
            out = ATOM + "(" + u + FALSE + ")"
            return out, u + FALSE
        elif atom[0] == "atom":
            return ATOM + "(" + ast2str(atom[1]) + ")", ast2str(atom[1])
        elif atom[0] == "cmp":
            out  = CMP + "(\"" + atom[1][1] + "\"," + ast2str(atom[1][0]) + ","
            out += ast2str(atom[1][2]) + ")"
            return out, ast2str(atom[1])
        else:
            raise utils.FatalException


    def __bf2str(self, bf):
        if bf[0] == "ext_atom":
            atom_reified, atom = self.__translate_ext_atom(bf[1])
            # fills self.atoms_in_bf
            self.atoms_in_bf.add((atom_reified, atom)) 
            return atom_reified
        elif bf[0] == "csp":
            self.atoms_in_bf.add((bf[1], ast2str(bf[2])))
            return bf[1]
        elif bf[0] == "neg":
            return NEG + "("  + self.__bf2str(bf[1][0]) + ")"
        elif bf[0] == "and":
            return AND+ "("  + self.__bf2str(
                   bf[1][0]) + "," + self.__bf2str(bf[1][1]) + ")"
        elif bf[0] == "or":
            return OR + "(" + self.__bf2str(
                   bf[1][0]) + "," + self.__bf2str(bf[1][1]) + ")"


    def __translate_lit(self, lit):
        neg  = 0
        while lit[0] == "neg" and neg<=1:
            neg += 1
            lit  = lit[1][0]
        if lit[0] == "ext_atom":
            atom_reified, atom = self.__translate_ext_atom(lit[1])
            return ("lit", (neg*(NEG+"(")) + atom_reified + (neg*")"),
                   (neg*(NOT+" ")) + atom)
        return None


    def __translate_bf(self, bf):
        out = self.__translate_lit(bf)
        if out is not None: 
            return out
        string = self.__bf2str(bf) # fills self.atoms_in_bf
        return ("bf", string, Statement.underscores + SAT + "(" + string +")")


    #
    # WARNING: must be called in the first str_ function (now str_body)
    #
    # modifies elements of self.body with triples
    #   (type,reified,string)
    # where:
    #   - type may be bf or lit
    #   - reified is what will be used to write for()
    #   - string  is what will be used to generate holds/2
    # it also fills self.atoms_in_bf with the atoms appearing in the body
    # and     fills self.bf     with the reified version of the body
    #
    def __analyze_body(self):
        # translate body
        for i in range(len(self.body)):
            # fills self.atoms_in_bf
            self.body[i] = self.__translate_bf(self.body[i]) 
        # fill self.bf 
        self.bf  = "".join([AND + "(" + x[1] + "," for x in self.body[:-1]])
        self.bf += self.body[-1][1]
        self.bf += "".join([                   ")" for x in self.body[:-1]])
        self.analyzed = True


    #
    # FUNCTIONS RETURNING STRINGS
    #


    # return the weight
    def str_weight(self):
        return "(" + ast2str(self.weight) + ")"


    # return the body with for() or name()
    # analyzes the body
    def str_body(self):
        if self.naming: 
            return NAME + "({})".format(ast2str(self.body))
        self.__analyze_body()
        return FOR + "({})".format(self.bf)


    # return rules for holds/2
    def str_holds(self, body):
        if self.naming:
            return ""
        if body != "":
            body = ", " + body
        head = Statement.underscores + HOLDS + "(" + str(self.bf) + ",0) :- "
        return head + ", ".join([x[2] for x in self.body]) + body + ".\n"


    # return rules for bf/1 with the boolean formulas which are not literals
    # sets PStatement.bfs to True when necessary
    def str_bf(self, body):
        if self.naming: 
            return ""
        if body != "": 
            body = " :- " + body
        init = Statement.underscores + BF + "(" 
        bfs = [init + x[1] + ")" + body + "." for x in self.body if x[0]=="bf"]
        if bfs != []:
            PStatement.bfs = True
            return "\n".join(bfs)+"\n"
        return ""


    # return rules for sat/1 with extended atoms appearing in
    # boolean formulas which are not literals
    def str_sat(self, body):
        if self.naming:
            return ""
        if len(self.atoms_in_bf) == 0:
            return ""
        if body != "":
            body = ", " + body
        init = Statement.underscores + SAT + "("
        end  = body + "."
        list = [init + x[0] + ") :- " + x[1] + end for x in self.atoms_in_bf]
        return "\n".join(list) + "\n"


    #
    # FUNCTIONS FOR GENERATING dom() AND preference_dom() ATOMS
    #

    def __get_arity_and_vars_termvec(self, termvec):
        # basic case
        if termvec == None or len(termvec)==0: 
            return 0, False
        # move to ntermvec
        ntermvec, arity, _has_vars = termvec, 1, False
        # iterate
        while len(ntermvec)==3:
            arity   += 1
            if not _has_vars:
                _has_vars = has_vars(ntermvec[2])
            ntermvec =  ntermvec[0]
        if not _has_vars: 
            _has_vars = has_vars(ntermvec[0])
        # return
        return arity, _has_vars


    def __get_signature_atom(self,atom):
        # strong negation?
        if atom[0] == "-":
            name = "-"
            atom = atom[1:]
        else: 
            name = ""
        # name
        name += atom[0]
        # if arity is 0
        if len(atom)==1:
            return None
        # else get arity, and whether are there variables, and return
        arity, _has_vars = self.__get_arity_and_vars_termvec(atom[2][0])
        return (name, arity) if _has_vars else None


    def __get_dom_atoms_from_bf(self,bf):
        if bf[0] == "ext_atom":
            if bf[1][0] == "atom":
                # (name,arity) for non ground atoms
                sig = self.__get_signature_atom(bf[1][1]) 
                if sig is None: 
                    return []
                u = Statement.underscores
                out1 = u +     DOM + "(" +          ast2str(bf[1][1]) + ")"
                out2 = u + GEN_DOM + "(" + sig[0] + "," + str(sig[1]) + ")."
                return [(out1, out2)]
            else: 
                return []
        elif bf[0] == "and" or bf[0] == "or":
            out1 =        self.__get_dom_atoms_from_bf(bf[1][0]) 
            return out1 + self.__get_dom_atoms_from_bf(bf[1][1])
        elif bf[0] == "neg":
            return self.__get_dom_atoms_from_bf(bf[1][0])
        else: 
            raise Exception("ERROR at __get_dom_atoms_from_bf")


    def get_dom_atoms(self):
        if self.naming:
            sig = self.__get_signature_atom(self.body)
            if sig is not None:
                init = Statement.underscores + PREFERENCE_DOM
                return [(init + "(" + ast2str(self.body) + ")", None)]
            else:
                return []
        return [item for list in map(self.__get_dom_atoms_from_bf,
                                     self.body) for item in list]


